import sys
import time
from contextlib import contextmanager
from pathlib import Path
from typing import Any, List, Optional, Tuple

import click
import typer
from typer.core import TyperGroup

from wheel2deb import logger as logging
from wheel2deb.build import build_all_packages, build_packages
from wheel2deb.context import load_configuration
from wheel2deb.debian import convert_wheels
from wheel2deb.logger import enable_debug
from wheel2deb.version import __version__

logger = logging.getLogger(__name__)


class DefaultCommandGroup(TyperGroup):
    """
    Make it so that calling wheel2deb without a subcommand
    is equivalent to calling the default subcommand.
    """

    def __init__(self, *args: Any, **kwargs: Any):
        self.default_command = "default"
        self.ignore_unknown_options = True
        super().__init__(*args, **kwargs)

    def parse_args(self, ctx: click.Context, args: List[str]) -> List[str]:
        if not args:
            args.insert(0, self.default_command)
        return super().parse_args(ctx, args)

    def get_command(self, ctx: click.Context, cmd_name: str) -> click.Command | None:
        if cmd_name.startswith("-") and cmd_name not in self.commands:
            cmd_name = self.default_command
            ctx.default_command = True  # type: ignore
        return super().get_command(ctx, cmd_name)

    def resolve_command(
        self, ctx: click.Context, args: List[str]
    ) -> Tuple[str | None, click.Command | None, List[str]]:
        cmd_name, cmd, args = super().resolve_command(ctx, args)
        if hasattr(ctx, "default_command") and cmd_name:
            args.insert(0, cmd_name)
        return cmd_name, cmd, args


option_verbose: bool = typer.Option(
    False,
    "--verbose",
    "-v",
    envvar="WHEEL2DEB_VERBOSE",
    help="Enable more logs.",
    callback=lambda v: enable_debug(v),
)

option_configuration: Optional[Path] = typer.Option(
    None,
    "--config",
    "-c",
    envvar="WHEEL2DEB_CONFIG",
    help="Path to configuration file.",
)

option_output_directory: Path = typer.Option(
    "output",
    "--output-dir",
    "-o",
    envvar="WHEEL2DEB_OUTPUT_DIR",
    help="Directory where debian source packages are generated and built.",
)

option_include_wheels: Optional[List[str]] = typer.Option(
    None,
    "--include",
    "-i",
    envvar="WHEEL2DEB_INCLUDE_WHEELS",
    help="Only wheels with matching names will be converted",
)

option_exclude_wheels: Optional[List[str]] = typer.Option(
    None,
    "--exclude",
    "-e",
    envvar="WHEEL2DEB_EXCLUDE_WHEELS",
    help="Wheels with matching names will not be converted",
)

option_search_paths: List[Path] = typer.Option(
    [Path(".")],
    "--search-path",
    "-x",
    envvar="WHEEL2DEB_SEARCH_PATHS",
    help="Only blueprints with matching names will be taken into account",
)


option_workers_count: int = typer.Option(
    4,
    "--workers",
    "-w",
    envvar="WHEEL2DEB_WORKERS_COUNT",
    help="Max number of source packages to build in parallel",
)

option_force_build: bool = typer.Option(
    False, "--force", help="Build source package even if .deb already exists"
)

app = typer.Typer(cls=DefaultCommandGroup)


@contextmanager
def print_summary_and_exit():
    start_time = time.monotonic()
    yield
    logger.summary(
        f"\nWarnings: {logging.get_warning_counter()}. "
        f"Errors: {logging.get_error_counter()}. "
        f"Elapsed: {round(time.monotonic() - start_time, 3)}s."
    )
    # the return code is the number of errors
    sys.exit(logging.get_error_counter())


def filter_wheels(
    search_paths: List[Path],
    include_wheels: List[str] | None,
    exclude_wheels: List[str] | None,
) -> List[Path]:
    # list all python wheels in search paths
    files = []
    for path in [Path(path) for path in search_paths]:
        files.extend(path.glob("*.whl"))
    files = sorted(files, key=lambda x: x.name)

    filenames = [f.name for f in files]
    if not include_wheels:
        include_wheels = filenames

    # remove excluded wheels
    if exclude_wheels:
        include_wheels = list(filter(lambda x: x not in exclude_wheels, include_wheels))

    return [file for file in files if file.name in include_wheels]


@app.command(help="Generate and build source packages.")
def default(
    verbose: bool = option_verbose,
    configuration_path: Optional[Path] = option_configuration,
    output_directory: Path = option_output_directory,
    search_paths: List[Path] = option_search_paths,
    include_wheels: Optional[List[str]] = option_include_wheels,
    exclude_wheels: Optional[List[str]] = option_exclude_wheels,
    workers_count: int = option_workers_count,
    force_build: bool = option_force_build,
) -> None:
    with print_summary_and_exit():
        settings = load_configuration(configuration_path)
        wheel_paths = filter_wheels(search_paths, include_wheels, exclude_wheels)
        packages = convert_wheels(settings, output_directory, wheel_paths)
        build_packages([p.root for p in packages], workers_count, force_build)


@app.command(help="Convert wheels in search paths to debian source packages")
def convert(
    verbose: bool = option_verbose,
    configuration_path: Optional[Path] = option_configuration,
    output_directory: Path = option_output_directory,
    search_paths: List[Path] = option_search_paths,
    include_wheels: Optional[List[str]] = option_include_wheels,
    exclude_wheels: Optional[List[str]] = option_exclude_wheels,
) -> None:
    with print_summary_and_exit():
        settings = load_configuration(configuration_path)
        wheel_paths = filter_wheels(search_paths, include_wheels, exclude_wheels)
        convert_wheels(settings, output_directory, wheel_paths)


@app.command(help="Build debian packages from source packages.")
def build(
    verbose: bool = option_verbose,
    output_directory: Path = option_output_directory,
    workers_count: int = option_workers_count,
    force_build: bool = option_force_build,
) -> None:
    with print_summary_and_exit():
        build_all_packages(output_directory, workers_count, force_build)


@app.command(help="Output wheel2deb version.")
def version() -> None:
    typer.echo(__version__)


def main() -> None:
    app()



import re
from functools import lru_cache
from typing import Optional

import attr

from wheel2deb import logger as logging
from wheel2deb.utils import shell

# https://www.debian.org/doc/debian-policy/ch-controlfields.html#version
PACKAGE_VER_RE = re.compile(
    r"^(?:(?P<epoch>\d+):)?"
    r"(?P<version>(?:[\w\.~\-]+(?=-(?P<revision>[^-]+$)))|[\w\.~\-]+)"
)

APT_CACHE_MADISON_RE = re.compile(r"[^|]+\|([^|]+)\|[^|]+")

logger = logging.getLogger(__name__)

_cache = None


@attr.s(frozen=True)
class Package:
    # package name
    name = attr.ib(type=str)
    # upstream version
    version = attr.ib(type=str)
    # debian revision
    revision = attr.ib(type=str)
    epoch = attr.ib(type=str)

    @classmethod
    def factory(cls, name, pkg_version):
        g = PACKAGE_VER_RE.match(pkg_version).groupdict()
        return cls(name, **g)

    def __str__(self):
        # show only package name and upstream version
        return "{}=={}".format(self.name, self.version)


@lru_cache
def search_package(name, arch) -> Optional[Package]:
    name = name + ":" + arch if arch else name
    output, _ = shell(["apt-cache", "madison", name])
    match = APT_CACHE_MADISON_RE.match(output)
    return Package.factory(name, match.group(1).strip()) if match is not None else None


def search_packages(names, arch):
    if not names:
        return

    logger.debug(f"searching {' '.join(names)} in apt cache...")

    for name in names:
        yield search_package(name, arch)



import re

from packaging.version import parse

from wheel2deb import logger as logging
from wheel2deb.apt import search_packages

logger = logging.getLogger(__name__)

APT_FILE_RE = re.compile(r"(.*lib.+):\s(?:/usr/lib/|/lib/)")

DEB_VERS_OPS = {
    "==": ">=",
    "<": "<<",
    ">": ">>",
    "~=": ">=",
}


def normalize_package_version(python_package_version, prerelease_workaround=True):
    """
    Normalize Python package version to be used as Debian package version.
    :param python_package_version: The version of a Python package (a string).
    :param prerelease_workaround: :data:`True` to enable the pre-release
                                  handling documented below, :data:`False` to
                                  restore the old behavior.
    Reformats Python package versions to comply with the Debian policy manual.
    All characters except alphanumerics, dot (``.``) and plus (``+``) are
    replaced with dashes (``-``).
    The PEP 440 pre-release identifiers 'a', 'b', 'c' and 'rc' are prefixed by
    a tilde (``~``) to replicate the intended ordering in Debian versions, also
    the identifier 'c' is translated into 'rc'. Refer to `issue #8
    <https://github.com/paylogic/py2deb/issues/8>`_ for details.
    """
    # Credits to https://github.com/paylogic/py2deb/blob/master/py2deb/utils.py
    # Lowercase and remove invalid characters from the version string.
    # Points (".") are stripped as well, since we end up with a trailing
    # point with versions like "7.*"
    version = (
        re.sub("[^a-z0-9.+]+", "-", python_package_version.lower()).strip("-").strip(".")
    )
    if prerelease_workaround:
        # Translate the PEP 440 pre-release identifier 'c' to 'rc'.
        version = re.sub(r"(\d)c(\d)", r"\1rc\2", version)
        # Replicate the intended ordering of PEP 440 pre-release versions (a, b, rc).
        version = re.sub(r"(\d)(a|b|rc)(\d)", r"\1~\2\3", version)
    return version


def suggest_name(ctx, wheel_name):
    """
    Guess Debian package name from a wheel name and a python implementation.
    :param wheel_name: Name of the distribution wheel
    :type wheel_name: str
    :return: Package name
    """

    prefix = {2: "python", 3: "python3"}[ctx.python_version.major]

    if wheel_name in ctx.map:
        return prefix + "-" + ctx.map[wheel_name]

    basename = re.compile("[^A-Za-z0-9.]+").sub("-", wheel_name)
    basename = basename.replace("python-", "")
    basename = basename.replace("-python", "").lower()

    return prefix + "-" + basename


def suggest_names(ctx, wheel_names):
    for wheel_name in wheel_names:
        yield suggest_name(ctx, wheel_name)


def search_python_deps(ctx, wheel, extras=None):
    """
    Search debian python dependencies
    :param wheel: Python wheel to guess dependencies for
    :param extras: List of wheels. Dependencies provided by this list will be
    considered satisfied
    :return: list of debian packages
    """

    extras = extras or []

    # keep only requirements that match the environment
    # https://www.python.org/dev/peps/pep-0508/#environment-markers
    requirements = wheel.requires(
        {
            "platform_machine": ctx.platform_machine,
            "python_version": str(ctx.python_version),
            "extra": ctx.extra,
        }
    )

    # filter out ignored requirements
    def is_required(r):
        if r.name in ctx.ignore_requirements:
            logger.warning(f"ignoring requirement {str(r)}")
            return False
        else:
            return True

    # remove ignored requirements
    requirements = list(filter(is_required, requirements))

    # remove duplicates
    requirements = list(set(requirements))

    # translate requirements to debian package names
    # and search them in apt cache
    debnames = list(suggest_names(ctx, [r.name for r in requirements]))
    results = search_packages(debnames, ctx.arch)

    candidates = {r.name: [] for r in requirements}
    for res, req in zip(results, requirements):
        if res:
            # add debian package to candidates list
            candidates[req.name].append(res)
        for extra in extras:
            if extra.name == req.name and extra.version_supported(ctx.python_version):
                # add extra wheel to candidates list
                candidates[extra.name].append(extra)

    debian_deps = []
    missing_deps = []
    for pdep, req in zip(debnames, requirements):

        def check(x):
            if req.specifier.contains(x.version, prereleases=True):
                logger.info(f"{x} satisfies requirement {req}")
            else:
                logger.warning(f"{x} does not satisfy requirement {req}")
            return ctx.ignore_upstream_versions or req.specifier.contains(
                x.version, prereleases=True
            )

        version = None
        for candidate in candidates[req.name]:
            if check(candidate):
                if (version and parse(candidate.version) < parse(version)) or not version:
                    version = candidate.version

        if not version:
            logger.error(f"could not find a candidate for requirement {req}")
            missing_deps.append(str(req))

        if req.name in ctx.ignore_specifiers:
            logger.warning(f"ignoring specifiers for dependency {req.name}")
            debian_deps.append(pdep)
        elif not ctx.ignore_upstream_versions and len(req.specifier):
            for specifier in req.specifier:
                dep = get_dependency_string(pdep, specifier.operator, specifier.version)
                if dep is not None:
                    debian_deps.append(dep)
        else:
            debian_deps.append(pdep)

    return debian_deps, missing_deps


def get_dependency_string(package_name, operator, version):
    """
    Gets a dependency string for a package based on an operator and version.
    """
    # != can't be translated to a package relationship in debian...
    if operator != "!=":
        v = normalize_package_version(version)
        if operator == "==":
            parsed = parse(v)
            if len(parsed.release) == 1:
                # For versions of the form "x" or "x.*". This will return "{x+1}.0"
                return f"{package_name} (<< {parsed.release[0] + 1})"
            elif len(parsed.release) > 1:
                # For version of the form "x.y" or "x.y.z". This will return "x.{y+1}"
                return f"{package_name} (<< {parsed.release[0]}.{parsed.release[1] + 1})"
        if operator == "<=":
            v += "-+"
        return f"{package_name} ({_translate_op(operator)} {v})"

    return None


def _translate_op(operator):
    """
    Translate Python version operator into Debian one.
    """
    return DEB_VERS_OPS.get(operator, operator)



import re
from pathlib import Path
from threading import Event, Thread
from time import sleep
from typing import List

from wheel2deb import logger as logging
from wheel2deb.utils import shell

logger = logging.getLogger(__name__)


def parse_debian_control(cwd: Path):
    """
    Extract fields from debian/control
    :param cwd: Path to debian source package
    :return: Dict object with fields as keys
    """

    field_re = re.compile(r"^([\w-]+)\s*:\s*(.+)")

    content = (cwd / "debian" / "control").read_text()
    control = {}
    for line in content.split("\n"):
        m = field_re.search(line)
        if m:
            g = m.groups()
            control[g[0]] = g[1]

    for k in ("Build-Depends", "Depends"):
        m = re.findall(r"([^=\s,()]+)\s?(?:\([^)]+\))?", control[k])
        control[k] = m

    return control


def build_package(cwd: Path) -> int:
    """Run dpkg-buildpackage in specified path."""
    args = ["dpkg-buildpackage", "-us", "-uc"]
    arch = parse_debian_control(cwd)["Architecture"]
    if arch != "all":
        args += ["--host-arch", arch]

    stdout, returncode = shell(args, cwd=cwd)
    logger.debug(stdout)
    if returncode:
        logger.error(f'failed to build package in "{cwd}" ☹')

    return returncode


def build_packages(paths: List[Path], threads: int, force_build: bool) -> None:
    """
    Run several instances of dpkg-buildpackage in parallel.
    :param paths: List of paths where dpkg-buildpackage will be called
    :param threads: Number of threads to run in parallel
    """

    paths = [p for p in paths if not Path(str(p) + ".deb").is_file() or force_build]
    logger.task(f"Building {len(paths)} source packages...")

    workers = []
    for i in range(threads):
        event = Event()
        event.set()
        workers.append({"done": event, "path": None})

    def build(done, path):
        logger.info(f"building {path}")
        build_package(path)
        done.set()

    while False in [w["done"].is_set() for w in workers] or paths:
        for w in workers:
            if w["done"].is_set() and paths:
                w["done"].clear()
                w["path"] = paths.pop()
                Thread(target=build, kwargs=w).start()
        sleep(1)


def build_all_packages(output_directory: Path, workers: int, force_build: bool) -> None:
    """
    Build debian source packages in parallel.
    :param output_directory: path where to search for source packages
    :param workers: Number of threads to run in parallel
    :param force_build: Build packages even if .deb already exists
    """

    if output_directory.exists() is False:
        logger.error(f"Directory {output_directory} does not exist")
        return

    if output_directory.is_dir() is False:
        logger.error(f"{output_directory} is not a directory")
        return

    paths = []
    for output_directory in output_directory.iterdir():
        if output_directory.is_dir() and (output_directory / "debian/control").is_file():
            paths.append(output_directory)

    build_packages(paths, workers, force_build)



import os
import subprocess
from pathlib import Path
from typing import List, Tuple


def shell(args: List[str], cwd: Path | None = None) -> Tuple[str, int]:
    env = os.environ.copy()
    env.pop("LD_LIBRARY_PATH", None)
    result = subprocess.run(
        args, cwd=cwd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, env=env
    )
    return result.stdout.decode("utf-8"), result.returncode



#  Copyright (c) 2015-2018 Cisco Systems, Inc.
#  Copyright (c) 2019 Parkoview SA
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.

import logging
import sys
from logging import DEBUG, ERROR, INFO, WARNING

import colorama

colorama.init(autoreset=True)

SUMMARY = 100
TASK = 101


class LogFilter(object):
    """
    A custom log filter which excludes log messages above the logged
    level.
    """

    def __init__(self, level):
        self.__level = level

    def filter(self, logRecord):  # pragma: no cover
        # https://docs.python.org/3/library/logging.html#logrecord-attributes
        return logRecord.levelno <= self.__level


class CustomLogger(logging.getLoggerClass()):
    """
    A custom logging class which adds additional methods to the logger.
    These methods serve as syntactic sugar for formatting log messages.
    """

    def __init__(self, name, level=logging.NOTSET):
        super(logging.getLoggerClass(), self).__init__(name, level)
        logging.addLevelName(SUMMARY, "SUCCESS")
        logging.addLevelName(TASK, "TASK")

    def summary(self, msg, *args, **kwargs):
        if self.isEnabledFor(SUMMARY):
            self._log(SUMMARY, msg, args, **kwargs)

    def task(self, msg, *args, **kwargs):
        if self.isEnabledFor(TASK):
            self._log(TASK, msg, args, **kwargs)


class CounterStreamHandler(logging.StreamHandler):
    """
    A StreamHandler which stores a call counter for each level
    Not thread-safe !
    """

    counters = {}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def emit(self, record):
        level = record.levelno
        if level not in self.counters:
            self.counters[level] = 0
        self.counters[level] += 1
        super().emit(record)


class TrailingNewlineFormatter(logging.Formatter):
    """
    A custom logging formatter which removes additional newlines from messages.
    """

    def format(self, record):
        if record.msg:
            record.msg = record.msg.rstrip()
        return super(TrailingNewlineFormatter, self).format(record)


class MultilineFormatter(logging.Formatter):
    """
    A custom logging formatter which formats every line in the record
    """

    def format(self, record):
        message = record.msg
        output = ""
        for line in message.splitlines():
            record.msg = line
            output += super().format(record) + "\n"
        record.msg = message
        record.message = output
        return output[:-2]


def get_warning_counter():
    if WARNING in CounterStreamHandler.counters:
        return CounterStreamHandler.counters[WARNING]
    else:
        return 0


def get_error_counter():
    if ERROR in CounterStreamHandler.counters:
        return CounterStreamHandler.counters[ERROR]
    else:
        return 0


def getLogger(name=None):
    """
    Build a logger with the given name and returns the logger.

    :param name: The name for the logger. This is usually the module
                 name, ``__name__``.
    :return: logger object
    """
    logging.setLoggerClass(CustomLogger)

    logger = logging.getLogger(name)
    logger.setLevel(logging.NOTSET)

    logger.addHandler(_get_debug_handler())
    logger.addHandler(_get_info_handler())
    logger.addHandler(_get_task_handler())
    logger.addHandler(_get_warn_handler())
    logger.addHandler(_get_error_handler())
    logger.addHandler(_get_critical_handler())
    logger.addHandler(_get_summary_handler())
    logger.propagate = False

    return logger


def _get_debug_handler():
    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(MultilineFormatter(dim_text("     %(message)s")))
    handler.setLevel(logging.DEBUG)
    handler.addFilter(LogFilter(logging.DEBUG))

    return handler


def _get_info_handler():
    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(TrailingNewlineFormatter("     %(message)s"))
    handler.setLevel(logging.INFO)
    handler.addFilter(LogFilter(logging.INFO))

    return handler


def _get_task_handler():
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(TASK)
    handler.addFilter(LogFilter(TASK))
    handler.setFormatter(
        TrailingNewlineFormatter(" --> {}".format(cyan_text("%(message)s")))
    )

    return handler


def _get_warn_handler():
    handler = CounterStreamHandler(sys.stdout)
    handler.setLevel(logging.WARN)
    handler.addFilter(LogFilter(logging.WARN))
    handler.setFormatter(TrailingNewlineFormatter(yellow_text("     %(message)s")))

    return handler


def _get_error_handler():
    handler = CounterStreamHandler(sys.stderr)
    handler.setLevel(logging.ERROR)
    handler.addFilter(LogFilter(logging.ERROR))
    handler.setFormatter(TrailingNewlineFormatter(red_text("     %(message)s")))

    return handler


def _get_critical_handler():
    handler = logging.StreamHandler(sys.stderr)
    handler.setLevel(logging.CRITICAL)
    handler.addFilter(LogFilter(logging.CRITICAL))
    handler.setFormatter(TrailingNewlineFormatter(red_text("%(message)s")))

    return handler


def _get_summary_handler():
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(SUMMARY)
    handler.addFilter(LogFilter(SUMMARY))
    handler.setFormatter(TrailingNewlineFormatter(green_text("     %(message)s")))

    return handler


def dim_text(msg):
    return color_text(colorama.Style.DIM, msg)


def red_text(msg):
    return color_text(colorama.Fore.RED, msg)


def yellow_text(msg):
    return color_text(colorama.Fore.YELLOW, msg)


def green_text(msg):
    return color_text(colorama.Fore.GREEN, msg)


def cyan_text(msg):
    return color_text(colorama.Fore.CYAN, msg)


def color_text(color, msg):
    return "{}{}{}".format(color, msg, colorama.Style.RESET_ALL)


def enable_debug(debug: bool):
    if debug:
        logging.getLogger().setLevel(DEBUG)
    else:
        logging.getLogger().setLevel(INFO)



from typing import Optional

from jinja2 import Environment, FunctionLoader

DEBIAN_CHANGELOG = """\
{{ package.name }} ({{ package.version }}) stable; urgency=medium

  * Release {{ package.version }}

 -- {{ ctx.maintainer_name }} <{{ ctx.maintainer_email }}>  Tue, 07 May 2019 20:31:30 +0000
"""

DEBIAN_COMPAT = """\
10
"""

DEBIAN_CONTROL = """\
Source: {{ package.name }}
Section: python
Priority: optional
Maintainer: {{ ctx.maintainer_name }} <{{ ctx.maintainer_email }}>
Build-Depends: debhelper
Standards-Version: 3.9.6

Package: {{ package.name }}
Architecture: {{ package.arch }}
Depends: {{ package.depends|sort|join(', ') }}
{%- if ctx.conflicts %}{{ '\n' }}Conflicts: {{ ctx.conflicts|sort|join(', ') }}{% endif %}
{%- if ctx.provides %}{{ '\n' }}Provides: {{ ctx.provides|sort|join(', ') }}{% endif %}
{%- if package.homepage %}{{ '\n' }}Homepage: {{ package.homepage }}{% endif %}
Description: {{ package.description }}
 {{ ctx.extended_desc }}
"""

DEBIAN_POSTINST = """\
#!/bin/sh
set -e

#WHEEL2DEB#

{% if package.pyvers.major == 2 %}
{% set pycompile = 'pycompile' %}
{% else %}
{% set pycompile = 'py3compile' %}
{% endif %}

if which {{ pycompile }} >/dev/null 2>&1; then
    {{ pycompile }} -p {{ package.name }}
fi

#DEBHELPER#
"""

DEBIAN_PRERM = """\
#!/bin/sh
set -e

#WHEEL2DEB#

{% if package.pyvers.major == 2 %}
{% set pyclean = 'pyclean' %}
{% else %}
{% set pyclean = 'py3clean' %}
{% endif %}

if which {{ pyclean }} >/dev/null 2>&1; then
    {{ pyclean }} -p {{ package.name }}
fi

#DEBHELPER#
"""

DEBIAN_COPYRIGHT = """\
Format: https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/

Files: *
Copyright: {{ copyrights|join('\n           ') }}
License: {{ license }}

License: {{ license }}
{{ license_content }}
"""

DEBIAN_RULES = """\
#!/usr/bin/make -f

%:
	dh $@

override_dh_shlibdeps:
	true

override_dh_builddeb:
	dh_builddeb -- -Zxz
"""

DEBIAN_ENTRYPOINT = """\
#!/usr/bin/python{{pyvers.major}}
from {{entrypoint.module}} import {{entrypoint.function}}
{{entrypoint.function}}()
"""


def template_loader(name: str) -> Optional[str]:
    variable_name = f"DEBIAN_{name.upper().replace('-', '_')}"
    template_content: str = globals()[variable_name]
    if variable_name in globals():
        return template_content
    return None


environment = Environment(loader=FunctionLoader(template_loader))



import re

import attr


@attr.s(frozen=True)
class Version:
    major = attr.ib(type=int, converter=int)
    minor = attr.ib(type=int, converter=int, default=0)
    micro = attr.ib(type=int, converter=int, default=0)

    @classmethod
    def from_str(cls, version_str):
        m = re.match(r"(\d)(?:\.(\d+))?(?:\.(\d+))?", version_str)
        v = [int(i) if i else 0 for i in m.groups()]
        return cls(v[0], v[1], v[2])

    def inc(self):
        return Version(self.major, self.minor + 1, 0)

    def __str__(self):
        return (
            str(self.major)
            + "."
            + str(self.minor)
            + (("." + str(self.micro)) if self.micro else "")
        )


@attr.s(frozen=True)
class VersionRange:
    min = attr.ib(type=Version, default=None)
    max = attr.ib(type=Version, default=None)

    @max.validator
    def check(self, attribute, value):
        """Enforce max > min. Interval must be open"""
        if value and value <= self.min:
            raise ValueError("min must be strictly smaller than max")

    def __contains__(self, item):
        if self.min and self.max:
            return self.min <= item < self.max
        elif self.min:
            return self.min <= item
        elif self.max:
            return item < self.max
        else:
            return True



import configparser
import os.path
import re
from dataclasses import dataclass
from functools import cached_property, lru_cache
from pathlib import Path
from typing import List

import attr
from packaging import specifiers, version
from packaging.requirements import Requirement
from pkginfo import Distribution
from wheel.wheelfile import WheelFile

from wheel2deb import logger as logging
from wheel2deb.pyvers import Version, VersionRange

logger = logging.getLogger(__name__)

EXTRACT_PATH = Path("/tmp/wheel2deb")

WHEEL_NAME_RE = re.compile(
    r"^(?P<name>.+)-(?P<version>.+)-(?P<python_tag>[pcij].+)"
    r"-(?P<abi_tag>.+)-(?P<platform_tag>.+).whl$"
)


def normalize_name(name):
    """
    All comparisons of distribution names MUST be case insensitive,
    and MUST consider hyphens and underscores to be equivalent.
    https://www.python.org/dev/peps/pep-0426/#name
    """
    return name.replace("-", "_").lower()


@dataclass
class Entrypoint:
    name: str
    module: str
    function: str


@attr.s(frozen=True)
class Record:
    """
    Entries of *.dist-info/RECORD organized in categories
    """

    LICENSE_RE = re.compile(r"(^|[^\w])license(\..*)?$", re.IGNORECASE)
    SHLIBS_RE = re.compile(r"\.so[.\d]*")

    libs = attr.ib(factory=list)
    lib_dirs = attr.ib(factory=list)
    licenses = attr.ib(factory=list)
    scripts = attr.ib(factory=list)
    files = attr.ib(factory=list)

    @classmethod
    def from_str(cls, content):
        files = [line.rstrip().split(",")[0] for line in content.split("\n")]
        record = Record()
        for file in files:
            if re.search(cls.LICENSE_RE, file):
                logger.debug(f"found license: {file}")
                record.licenses.append(file)
                continue

            if ".data/scripts/" in file:
                logger.debug(f"found script: {file}")
                record.scripts.append(file)
                continue

            if re.findall(cls.SHLIBS_RE, os.path.basename(file)):
                logger.debug(f"found shared lib: {file}")
                record.libs.append(file)

            if file:
                # everything else
                record.files.append(file)

        record.lib_dirs.extend(list({os.path.dirname(x) for x in record.libs}))

        return record


class Metadata(Distribution):
    def __init__(self, content):
        self.parse(content)

    def read(self):
        pass


class Wheel:
    def __init__(self, wheel_name: str, extract_path: Path) -> None:
        self.wheel_name = wheel_name
        self.extract_path = extract_path
        self.info_dir = next(iter(self.extract_path.glob("*.dist-info")))

        # parse wheel name, see https://www.python.org/dev/peps/pep-0425
        g = re.match(WHEEL_NAME_RE, self.wheel_name).groupdict()
        self.name = normalize_name(g["name"])
        self.version = g["version"]
        self.python_tag = g["python_tag"]
        self.abi_tag = g["abi_tag"]
        self.platform_tag = g["platform_tag"]

    @cached_property
    def metadata(self) -> Metadata:
        return Metadata((self.info_dir / "METADATA").read_text())

    @cached_property
    def record(self) -> Record:
        return Record.from_str((self.info_dir / "RECORD").read_text())

    @cached_property
    def entrypoints(self) -> List[Entrypoint]:
        entrypoints = []
        try:
            config = configparser.ConfigParser()
            config.read_string((self.info_dir / "entry_points.txt").read_text())
            if "console_scripts" in config.sections():
                name, path = config.items("console_scripts")[0]
                entrypoints.append(Entrypoint(name, *(path.split(":"))))
        except FileNotFoundError:
            pass
        return entrypoints

    def requires(self, env=None):
        if not env:
            env = {"extra": ""}
        elif env and "extra" not in env:
            env["extra"] = ""
        reqs = [Requirement(req) for req in self.metadata.requires_dist]
        reqs = filter(lambda r: not r.marker or r.marker.evaluate(env), reqs)
        reqs = list(reqs)
        for req in reqs:
            req.name = normalize_name(req.name)
        return reqs

    @lru_cache
    def version_range(self, pyvers):
        m = re.search(r"(\d)(\d+)", self.python_tag)

        if m:
            v = Version(*m.groups())
            if pyvers.major != v.major:
                return None
            else:
                if self.abi_tag == "abi3":
                    return VersionRange(v, None)
                return VersionRange(v, v.inc())

        # TODO: use requires_python ?
        versions = []
        for classifier in self.metadata.classifiers:
            m = re.match(r"Programming Language :: Python :: ([\d.]+)", classifier)
            if m:
                version = Version.from_str(m.group(1))
                if version.major == pyvers.major and version.minor != 0:
                    versions.append(version)
        sorted(versions)

        if versions:
            # assume versions[0] and up supported
            return VersionRange(versions[0], None)

        # unable to compute python version range
        # supported by that wheel
        return None

    @lru_cache
    def version_supported(self, pyvers):
        m = re.search(r"(?:py|cp)%s" % pyvers.major, self.python_tag)
        if not m:
            return False

        requires_python = self.metadata.requires_python
        if self.metadata.requires_python is None:
            # The package provides no information
            return True

        requires_python_specifier = specifiers.SpecifierSet(requires_python)
        python_version = version.parse(str(pyvers))
        return python_version in requires_python_specifier

    @cached_property
    def cpython_supported(self):
        if re.search(r"(?:py|cp)", self.python_tag):
            return True
        return False

    def __repr__(self):
        return self.wheel_name


def parse_wheel(wheel_path: Path, base_extract_path: Path = EXTRACT_PATH) -> Wheel:
    extract_path = base_extract_path / wheel_path.name[:-4]
    if extract_path.exists() is False:
        with WheelFile(str(wheel_path)) as wf:
            logger.debug(f"unpacking wheel to: {extract_path}...")
            wf.extractall(str(extract_path))
    return Wheel(wheel_path.name, extract_path)



import re
from pathlib import Path
from typing import List

from dirsync import sync

from wheel2deb import logger as logging
from wheel2deb.context import Settings
from wheel2deb.depends import normalize_package_version, search_python_deps, suggest_name
from wheel2deb.pydist import Wheel, parse_wheel
from wheel2deb.templates import environment
from wheel2deb.utils import shell
from wheel2deb.version import __version__

logger = logging.getLogger(__name__)
dirsync_logger = logging.getLogger("dirsync")
dirsync_logger.setLevel(logging.ERROR)


COPYRIGHT_RE = re.compile(
    r"(?:copyrights?|\s*©|\s*\(c\))[\s:|,]*" r"((?=.*[a-z])\d{2,4}(?:(?!all\srights).)+)",
    re.IGNORECASE,
)

DPKG_SHLIBS_RE = re.compile(r"find library (.+\.so[.\d]*) needed")

APT_FILE_RE = re.compile(r"(.*lib.+):\s(?:/usr/lib/|/lib/)")


def platform_to_arch(platform_tag):
    translation_table = {
        "x86_64": "amd64",
        "i686": "i686",
        "armv7l": "armhf",
        "armv6l": "armhf",
        "aarch64": "arm64",
        "any": "all",
    }
    for k, v in translation_table.items():
        if k in platform_tag:
            return v
    return None


class SourcePackage:
    """
    Create a debian source package that can be built
    with dpkg-buildpackage from a python wheel
    """

    def __init__(self, ctx, wheel: Wheel, output, extras=None):
        self.wheel = wheel
        self.ctx = ctx
        self.pyvers = ctx.python_version

        # debian package name
        self.name = suggest_name(ctx, wheel.name)

        # debian package version
        self.version = ctx.version_template.format(
            epoch=ctx.epoch,
            upstream_version=normalize_package_version(self.wheel.version),
            revision=ctx.revision,
            w2d_version=__version__,
        )

        # debian package homepage
        self.homepage = wheel.metadata.home_page
        # debian package description
        self.description = wheel.metadata.summary
        # debian package extended description
        self.extended_desc = ctx.extended_desc
        # upstream license
        self.license = wheel.metadata.license or "custom"

        # debian package architecture
        arch = platform_to_arch(wheel.platform_tag)
        self.arch = arch or "all"
        if not arch:
            logger.error("unknown platform tag, assuming arch=all")

        version_without_epoch = self.version.split(":")[-1]
        # debian package full filename
        self.filename = f"{self.name}_{version_without_epoch}_{self.arch}.deb"

        # root directory of the debian source package
        self.root = Path(output) / self.filename[:-4]
        # contains the files extracted from the wheel
        self.src = Path("src")
        # debian directory path
        # holds the package config files
        self.debian = self.root / "debian"

        # sync src directory with files from the wheel
        sync(
            str(wheel.extract_path),
            str(self.root / self.src),
            "sync",
            create=True,
            logger=dirsync_logger,
        )

        self.interpreter = "python" if self.pyvers.major == 2 else "python3"

        # compute package run dependencies
        self.depends = [f"{self.interpreter}:any"]
        if vrange := wheel.version_range(self.pyvers):
            if vrange.max:
                self.depends.append(f"{self.interpreter} (<< {vrange.max})")
            if vrange.min:
                self.depends.append(f"{self.interpreter} (>= {vrange.min}~)")

        deps, missing = search_python_deps(ctx, wheel, extras)
        self.depends.extend(deps)
        self.depends.extend(ctx.depends)

        # write unsatisfied requirements in missing.txt
        (self.root / "missing.txt").write_text("\n".join(missing) + "\n")

    def install_console_scripts(self) -> None:
        output_path = self.root / "entrypoints"
        output_path.mkdir(exist_ok=True)
        for entrypoint in self.wheel.entrypoints:
            template = environment.get_template("entrypoint")
            script_path = str(output_path / entrypoint.name)
            template.stream(pyvers=self.pyvers, entrypoint=entrypoint).dump(script_path)

    def install(self):
        """Generate debian/install"""

        install = set()

        # wheel modules install path
        if self.pyvers.major == 2:
            destination_path = "/usr/lib/python2.7/dist-packages/"
        else:
            destination_path = "/usr/lib/python3/dist-packages/"

        for absolute_source_path in (self.root / self.src).iterdir():
            source_path = absolute_source_path.relative_to(self.root)
            if not source_path.name.endswith(".data"):
                install.add(f"{source_path} {destination_path}")
            else:
                if (absolute_source_path / "purelib").exists():
                    install.add(f"{source_path / 'purelib/*'} {destination_path}")

        if self.wheel.entrypoints and not self.ctx.ignore_entry_points:
            self.install_console_scripts()
            install.add("entrypoints/* /usr/bin/")

        for script in self.wheel.record.scripts:
            install.add(f"{self.src / script} /usr/bin")

        (self.debian / "install").write_text("\n".join(install))

    def rules(self):
        """Generate debian/rules"""

        self.dump_template(
            "rules",
            shlibdeps_params="".join(
                [" -l" + str(self.src / x) for x in self.wheel.record.lib_dirs]
            ),
        )

    def copyright(self):
        """
        Generate debian/copyright
        https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/
        """
        licenses = self.wheel.record.licenses
        license_file = None
        license_content = ""
        copyrights = set()

        if not licenses:
            logger.warning("no license found !")
            return

        # gather copyrights from all licenses
        for lic in licenses:
            content = (self.wheel.extract_path / lic).read_text()
            copyrights.update(set(re.findall(COPYRIGHT_RE, content)))

        copyrights = sorted(copyrights)

        logger.debug("found the following copyrights: %s", copyrights)

        for file in licenses:
            if "dist-info" in file:
                license_file = file
        if not license_file:
            license_file = licenses[0]

        with (self.wheel.extract_path / license_file).open() as f:
            for line in f.readlines():
                license_content += " " + line

        if license_content:
            self.dump_template(
                "copyright",
                license=self.license,
                license_content=license_content,
                copyrights=copyrights,
            )
        else:
            logger.warning("license not found !")

        # FIXME: licenses should not be copied by the install script

    def create(self):
        if not self.debian.exists():
            self.debian.mkdir(parents=True)

        for template in [
            "changelog",
            "control",
            "compat",
            "postinst",
            "prerm",
        ]:
            self.dump_template(template)

        self.fix_shebangs()
        self.rules()
        self.install()
        self.copyright()

        # dpkg-shlibdeps won't work without debian/control
        self.search_shlibs_deps()

        # re-generate debian/control with deps found by dpkg-shlibdeps
        self.dump_template("control")

    def dump_template(self, template_name, **kwargs):
        template = environment.get_template(template_name)
        output_path = str(self.debian / template_name)
        template.stream(package=self, ctx=self.ctx, **kwargs).dump(output_path)

    def fix_shebangs(self):
        files = [self.root / self.src / x for x in self.wheel.record.scripts]
        for file in files:
            content = file.read_text()
            shebang = "#!/usr/bin/env python%s" % self.ctx.python_version.major
            if not content.startswith(shebang):
                content = content.split("\n")
                content[0] = shebang
                content = "\n".join(content)
                with file.open("w") as g:
                    g.write(content)

    def search_shlibs_deps(self):
        """
        Search packages providing shared libs dependencies
        :return: List of packages providing those libs
        """
        shlibdeps = set()
        missing_libs = set()

        if self.wheel.record.lib_dirs:
            args = (
                ["dpkg-shlibdeps"]
                + ["-l" + str(self.src / x) for x in self.wheel.record.lib_dirs]
                + [str(self.src / x) for x in self.wheel.record.libs]
            )
            output, _ = shell(args, cwd=self.root)
            missing_libs.update(DPKG_SHLIBS_RE.findall(output, re.MULTILINE))

        if missing_libs:
            logger.info(
                "dpkg-shlibdeps reported the following missing "
                "shared libs dependencies: %s",
                missing_libs,
            )

            # search packages providing those libs
            for lib in missing_libs:
                output, _ = shell(["apt-file", "search", lib, "-a", self.arch])
                packages = set(APT_FILE_RE.findall(output))

                # remove dbg packages
                packages = [p for p in packages if p[-3:] != "dbg"]

                if not len(packages):
                    logger.warning("did not find a package providing %s", lib)
                else:
                    # we pick the package with the shortest name
                    packages = sorted(packages, key=len)
                    shlibdeps.add(packages[0])

                if len(packages) > 1:
                    logger.warning(
                        f"several packages providing {lib}: {packages}, picking "
                        f"{packages[0]}, edit debian/control to use another one."
                    )

        if shlibdeps:
            logger.info(f"detected dependencies: {shlibdeps}")

        self.depends = list(set(self.depends) | shlibdeps)


def convert_wheels(
    settings: Settings,
    output_directory: Path,
    wheel_paths: List[Path],
) -> List[SourcePackage]:
    if output_directory.exists() is True and output_directory.is_dir() is False:
        logger.error(f"{output_directory} is not a directory")
        return []

    output_directory.mkdir(exist_ok=True, parents=True)

    if wheel_paths:
        logger.task("Unpacking %s wheels", len(wheel_paths))

    wheels = []
    for file in wheel_paths:
        wheel = parse_wheel(file)
        ctx = settings.get_ctx(wheel.wheel_name)

        if not wheel.cpython_supported:
            # ignore wheels that are not cpython compatible
            logger.warning(f"{wheel.wheel_name} does not support cpython")
            continue

        if not wheel.version_supported(ctx.python_version):
            # ignore wheels that are not compatible specified python version
            logger.warning(
                f"{wheel.wheel_name} does not support python {ctx.python_version}"
            )
            continue

        logger.info("%s", wheel.wheel_name)
        wheels.append(wheel)

    packages = []
    for wheel in wheels:
        logger.task(f"Converting wheel {wheel}")
        ctx = settings.get_ctx(wheel.wheel_name)
        package = SourcePackage(ctx, wheel, output_directory, extras=wheels)
        package.create()
        packages.append(package)

    return packages



import platform
import re
import sys
from pathlib import Path

import attr
import yaml

from wheel2deb import logger as logging
from wheel2deb.pyvers import Version
from wheel2deb.version import __version__

logger = logging.getLogger(__name__)


@attr.s
class Context:
    maintainer_name = attr.ib(default="wheel2deb")
    maintainer_email = attr.ib(default="wheel2deb@upciti.com")
    distribution = attr.ib(default="unstable")
    python_version = attr.ib(
        converter=lambda x: Version.from_str(x) if isinstance(x, str) else x,
        on_setattr=attr.setters.convert,
        default=platform.python_version(),
    )
    platform_machine = attr.ib(default=platform.machine())
    arch = attr.ib(default="")
    ignore_entry_points = attr.ib(default=False)
    ignore_upstream_versions = attr.ib(default=False)
    ignore_requirements = attr.ib(factory=list)
    ignore_specifiers = attr.ib(factory=list)
    extra = attr.ib(default="")
    map = attr.ib(factory=dict)
    depends = attr.ib(factory=list)
    conflicts = attr.ib(factory=list)
    provides = attr.ib(factory=list)
    revision = attr.ib(default="1")
    epoch = attr.ib(default=0, converter=int)
    version_template = attr.ib(
        default="{epoch}:{upstream_version}-{revision}~w2d{w2d_version[0]}"
    )
    extended_desc = attr.ib(
        default="This package was generated by wheel2deb v" + str(__version__)
    )

    def update(self, changes):
        for k, v in changes.items():
            if v and hasattr(self, k):
                setattr(self, k, changes[k])


@attr.s
class Settings:
    config = attr.ib(factory=dict)
    default_ctx = attr.ib(factory=Context)

    def get_ctx(self, key):
        ctx = self.default_ctx
        for k in self.config.keys():
            if re.match(k, key):
                ctx = attr.evolve(ctx, **self.config[k])
        return ctx


def load_configuration(configuration_path: Path | None) -> Settings:
    default_configuration_path = Path("wheel2deb.yml")

    if configuration_path is None and default_configuration_path.is_file():
        configuration_path = default_configuration_path

    if configuration_path is None:
        return Settings()

    if configuration_path.exists() is False:
        logger.error(f"Configuration {configuration_path} does not exist")
        sys.exit(1)

    if configuration_path.is_file() is False:
        logger.error(f"{configuration_path} is not a file")
        sys.exit(1)

    try:
        configuration = yaml.safe_load(configuration_path.read_text())
    except yaml.YAMLError as e:
        logger.error(f"Invalid YAML in configuration file: {e}")
        sys.exit(1)

    return Settings(configuration)
